import copy
import pytest

from dli.models.paginator import Paginator


@pytest.fixture
def empty_package_request():
    return {
            "class": ["Access Requests", "collection"],
            "properties": {"page": 1, "pages_count": 1},
            "entities": []
           }


class TestPaginator:
    def test_retrieve_zero_packages(self, test_client, empty_package_request,
                                    dataset_request_v2):
        test_client._session.get.return_value.json.side_effect = [
            empty_package_request, dataset_request_v2
        ]

        paginator = Paginator(
            '/',
            test_client._Package,
        )

        assert len(list(paginator)) == 0

    def test_threaded_paginator(self, test_client,
                                package_request_index_v2):

        # Note: there is no side-effect call to dataset because the call that was
        # previously made in the constructor by .shape has now been made lazy to
        # avoid an explosion of calls.
        test_client._session.get.return_value.json.side_effect = [
            package_request_index_v2]

        paginator = Paginator(
            '/',
            test_client._Package,
            lambda a, page_size: {}
        )

        assert len(list(paginator)) == 2




import uuid
from functools import partial

import pytest

from dli.client.exceptions import CatalogueEntityNotFoundException
from tests.common import SdkIntegrationTestCase


@pytest.mark.integration
class PackageTestCase(SdkIntegrationTestCase):

    def test_get_unknown_package_raises_package_not_found(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_package(name="unknown")


    @pytest.mark.xfail(reason="Need to figure out how to mock permissions")
    def test_can_get_package_by_id_or_name(self):
        package_id = self.create_package(
            name="test_get_package_returns_non_siren_response"
        )
        package = self.client.get_package(package_id)
        self.assertEqual(package.id, package_id)

        package_by_name = self.client.get_package(name=package.name)
        self.assertEqual(package_by_name.id, package_id)

    def test_cannot_get_package_without_package_id_or_name(self):
        with self.assertRaises(ValueError):
            self.client.get_package(None)

    def test_get_datasets_in_package(self):
        client = self.client
        package_id = self.create_package("test_get_datasets_in_package")
        builder = self.dataset_builder(package_id, "test_package_functions").with_external_storage("somewhere")
        self.client.register_dataset(builder)
        datasets = client.get_package_datasets(package_id)
        self.assertEqual(len(datasets), 1)


@pytest.mark.integration
class GetDefaultTermsAndConditionsTestCase(SdkIntegrationTestCase):
    _DEFAULT_TERMS_AND_CONDITIONS = ('By submitting this Data request and checking the "Accept Terms and Conditions" '
                                     'box, you acknowledge and agree to the following:\n'
                                     '\n'
                                     '* To promptly notify the relevant Access Manager/Producer of your intended use '
                                     'of the Data;\n'
                                     '* To obtain the terms and conditions relevant to such use for such Data from '
                                     'the Producer;\n'
                                     '* To distribute such terms and conditions to each member of your '
                                     'Consumer Group who may use the Data;\n'
                                     '* To use the Data solely for such intended use, subject to such terms and '
                                     'conditions;\n'
                                     '* To ensure that the Data is only accessed by members of your Consumer Group, '
                                     'and only used by such members for such intended use, subject to such terms and '
                                     'conditions;\n'
                                     '* To adhere to any additional requests of Producer with respect to the Data '
                                     '(including but not limited to ceasing use of the Data and deleting the Data, '
                                     'and ensuring other members of the Consumer Group do so, upon revocation of your '
                                     'license by Producer).\n'
                                     '\n'
                                     'Please refer to the <a href="/terms-of-use" target="_blank">EULA</a> for any '
                                     'defined terms used above. '
                                     'The <a href="/terms-of-use" target="_blank">EULA</a> '
                                     'is the document you agreed to adhere to by accessing the Lake.')

    def test_get_default_terms_and_conditions_returns_proper_text(self):
        from dli.client.components.package import Package

        result = Package.get_default_package_terms_and_conditions(
            organisation_name='some_organisation_name')
        assert result == ''

        default_text = Package.get_default_package_terms_and_conditions(
            organisation_name='IHS Markit')
        assert default_text == self._DEFAULT_TERMS_AND_CONDITIONS



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import multiprocessing
import warnings
from typing import List, Optional
from concurrent.futures.thread import ThreadPoolExecutor
from urllib.parse import urljoin
import humps  # noqa: I900

from dli.models.paginator import Paginator
from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import consumption_urls, dataset_urls
from dli.models import log_public_functions_calls_using, AttributesDict
from dli.models.file_model import FileModel

THREADS = multiprocessing.cpu_count()

class InstanceModel(AttributesDict):

    def __init__(self, **kwargs):
        # Ignore the datafile's files
        kwargs.pop('files', [])
        super().__init__(**kwargs)

    def files(self) -> List[FileModel]:
        """
        :return: list of file models for files in the instance.
        """
        url = urljoin(
            self._client._environment.consumption,
            consumption_urls.consumption_manifest.format(
                id=self.datafile_id
            )
        )

        response = self._client.session.get(url)
        return [
            self._client._File(
                datafile_id=self.datafile_id,
                **d['attributes']
            )
            for d in response.json()['data']
        ]

    @classmethod
    def _from_v1_entity(cls, entity, page_size=None):
        properties = humps.decamelize(entity['properties'])
        return cls(**properties)

    def download_all(self, to: Optional[str]='./', flatten: Optional[bool]=False):
        warnings.warn(
            'This method will soon become deprecated. '
            'Please use `download` function instead.',
            PendingDeprecationWarning
        )
        return self.download(to, flatten)

    def download(self, to: Optional[str]='./', flatten: Optional[bool]=False) -> List[str]:
        """
        Download the files for the instance, then return a list of the file
        paths that were written.

        :param str to: The path on the system, where the files
            should be saved. must be a directory, if doesn't exist, will be
            created.

        :param bool flatten: The flatten parameter default behaviour
            (flatten=False) will allow the user to specify that they would
            like to keep the underlying folder structure when writing the
            downloaded files to disk.
            When “flatten” is set to True the
            folder structure and download all data to the specified path in
            and retains only the original file name. This is not useful if the
            file names are identical for multiple files.

        :return: List of paths to the downloaded files.

        :example:

            Downloading without flatten:

            .. code-block:: python

                >>> dataset.instances.latest().download('./local/path/')
                [
                  'as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                  'as_of_date=2019-09-11/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
                ]


        :example:

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:

            .. code-block:: python

                >>> dataset.instances.latest().download('./local/path/', flatten=True)
                [
                  'StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                  'StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
                ]


        """
        files = self.files()

        def _download(file_: FileModel):
            return file_.download(to=to, flatten=flatten)

        threads = min(THREADS, max(1, len(files)))
        with ThreadPoolExecutor(max_workers=threads) as executor:
            return list(executor.map(_download, files))

    def __str__(self):
        return f"\nINSTANCE {self.datafile_id} (size: {self.total_size/1048576 :.2f} Mb)"


log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['datafile_id']
)(InstanceModel)


class InstancesCollection(AttributesDict):

    def __init__(self, dataset=None, page_size=25):
        self._dataset = dataset
        self._paginator = Paginator(
            dataset_urls.datafiles.format(id=self._dataset.id),
            self._client._Instance, self._client._Instance._from_v1_entity,
            page_size=page_size,
        )

    def latest(self) -> InstanceModel:
        """:return: The latest instance."""
        response = self._client.session.get(
            dataset_urls.latest_datafile.format(id=self._dataset.id)
        ).json()

        return self._client._Instance._from_v1_entity(response)

    def all(self) -> List[InstanceModel]:
        """:return: All the instances."""
        warnings.warn(
            'The result of calling `.all` will be cached. If you want fresh '
            'results the next time you call `.all`, then please re-create the '
            'dataset variable before calling `.all`.'
        )
        yield from self._paginator


log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['_dataset.dataset_id']
)(InstancesCollection)



from unittest.mock import MagicMock, call

import pytest
from s3transfer.futures import TransferFuture

from dli.models.unstructured_dataset_model import UnstructuredDatasetModel
from dli.models.structured_dataset_model import StructuredDatasetModel


class TestUnstructuredDataset:

    def __setup_mocks(
        self,
        monkeypatch,
    ):
        # This has a directory name starting with `.` so must be ignored.
        self.key_latest_0 = MagicMock()
        self.key_latest_0.key = 'abc/.latest/nulls_and_unicode.parquet'
        self.key_latest_0.size = 1

        self.key_document1_attachment_0 = MagicMock()
        self.key_document1_attachment_0.key = 'abc/document_id=1/attachment/0.zip'
        self.key_document1_attachment_0.size = 1
        self.key_document1_document_0 = MagicMock()
        self.key_document1_document_0.key = 'abc/document_id=1/document/0.xml'
        self.key_document1_document_0.size = 1
        self.key_document1_document_1 = MagicMock()
        self.key_document1_document_1.key = 'abc/document_id=1/document/0.parquet'
        self.key_document1_document_1.size = 1
        self.key_document1_metadata_0 = MagicMock()
        self.key_document1_metadata_0.key = 'abc/document_id=1/metadata/0.json'
        self.key_document1_metadata_0.size = 1

        self.key_document2_attachment_0 = MagicMock()
        self.key_document2_attachment_0.key = 'abc/document_id=2/attachment/0.zip'
        self.key_document2_attachment_0.size = 1
        self.key_document2_document_0 = MagicMock()
        self.key_document2_document_0.key = 'abc/document_id=2/document/0.xml'
        self.key_document2_document_0.size = 1
        self.key_document2_document_1 = MagicMock()
        self.key_document2_document_1.key = 'abc/document_id=2/document/0.parquet'
        self.key_document2_document_1.size = 1
        self.key_document2_metadata_0 = MagicMock()
        self.key_document2_metadata_0.key = 'abc/document_id=2/metadata/0.json'
        self.key_document2_metadata_0.size = 1

        self.filter_mock = MagicMock()
        self.filter_mock.side_effect = iter([
            [
                self.key_latest_0,
                self.key_document1_attachment_0,
                self.key_document1_document_0,
                self.key_document1_document_1,
                self.key_document1_metadata_0,
                self.key_document2_attachment_0,
                self.key_document2_document_0,
                self.key_document2_document_1,
                self.key_document2_metadata_0,
            ]
        ])

        self.objects_mock = MagicMock()
        self.objects_mock.return_value.filter = self.filter_mock

        self.bucket_mock = MagicMock()
        self.bucket_mock.return_value.objects = self.objects_mock()

        self.resource_mock = MagicMock()
        self.resource_mock.return_value.Bucket = self.bucket_mock

        self.resource_mock.return_value.meta.client.list_objects_v2.return_value = {
            'CommonPrefixes': [
                {'Prefix': 'abc/.latest/'},
                {'Prefix': 'abc/document_id=1/'},
                {'Prefix': 'abc/document_id=2/'},
            ]
        }

        self.boto_mock = MagicMock()
        self.boto_mock.return_value.resource.return_value = self.resource_mock()
        monkeypatch.setattr(
            'dli.models.dataset_model.create_refreshing_session',
            self.boto_mock
        )

        self.transfer_future_mock = MagicMock(return_value=TransferFuture())
        self.transfer_future_mock.return_value.result = True
        self.download_mock = MagicMock()
        self.download_mock.download = lambda x: self.transfer_future_mock
        monkeypatch.setattr(
            'dli.models.dataset_model.DatasetModel._get_transfer_manager',
            self.download_mock
        )

    def assert_calls(self):
        self.boto_mock.assert_called()
        self.resource_mock.assert_called_once()
        self.bucket_mock.assert_has_calls([
            call('abc')
        ])
        self.objects_mock.assert_called_once()
        self.filter_mock.assert_has_calls([
            # Assert that we only call the non-hidden prefixes.
            call(Prefix='abc/'),
        ])

    def test_model_factory_instantiation(self, test_client, unstructured_dataset_request_index_v2):
        test_client._session.get.return_value.json.side_effect = [
            unstructured_dataset_request_index_v2,
        ]

        test_ds = test_client.datasets()

        assert type(test_ds.get("TestDataset")).__name__ == UnstructuredDatasetModel.__name__
        assert type(test_ds.get("OtherDataset")).__name__ == StructuredDatasetModel.__name__

    def test_list_plain(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        ls = dataset.list()

        self.assert_calls()

        assert len(ls) == 9, f'List: {ls}'

    def test_list_method_with_structured_args(self, object_summaries_s3, monkeypatch, client):
        class MockResource:
            class Objects:
                def filter(self, Prefix):
                    return object_summaries_s3

            def Bucket(self, *args):
                return self

            @property
            def objects(self):
                objects = MockResource.Objects()
                return objects

        boto_mock = MagicMock()
        boto_mock.return_value.resource.return_value = MockResource()
        monkeypatch.setattr(
            'dli.models.dataset_model.create_refreshing_session',
            boto_mock
        )

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        with pytest.raises(AttributeError) as e_info:
            dataset.list(partitions=["document_id=1"])
        assert "EXISTS" in str(e_info.value.args[0])

    def test_none_existing_method(self, object_summaries_s3, monkeypatch, client):

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        with pytest.raises(AttributeError) as e_info:
            dataset.dask_dataframe()
        assert "EXISTS" not in str(e_info.value.args[0])

    def test_download_plain(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        ls = dataset.download("test")

        self.assert_calls()

        assert len(ls) == 8, f'List: {ls}'

    def test_download_method_with_structured_args(self, unstructured_object_summaries_s3, monkeypatch, client):
        class MockResource:
            name = 'some-bucket'

            class Objects:
                def filter(self, Prefix):
                    return unstructured_object_summaries_s3

            def Bucket(self, *args):
                return self

            @property
            def meta(self):
                return MagicMock()

            @property
            def objects(self):
                objects = MockResource.Objects()
                return objects

        boto_mock = MagicMock()
        boto_mock.return_value.resource.return_value = MockResource()
        monkeypatch.setattr(
            'dli.models.dataset_model.create_refreshing_session',
            boto_mock
        )

        mock = MagicMock(return_value=TransferFuture())
        mock.return_value.result = True
        download_mock = MagicMock()
        download_mock.download = lambda x: mock
        monkeypatch.setattr(
            'dli.models.dataset_model.DatasetModel._get_transfer_manager',
            download_mock
        )

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        with pytest.raises(AttributeError) as e_info:
            dataset.download("test", flatten=True, partitions="document_id=1")
        assert "EXISTS" in str(e_info.value.args[0])

    def test_download_filter_attachments_plain(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        ls = dataset.download("test", with_attachments=False)

        self.assert_calls()

        assert all(map(lambda x: "attachment" not in x, ls))
        assert len(ls) == 6

    def test_download_filter_metadata(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        ls = dataset.download("test", with_metadata=False)

        self.assert_calls()

        assert all(map(lambda x: "metadata" not in x, ls))
        assert len(ls) == 6

    def test_download_filter_attachments_and_metadata(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        ls = dataset.download("test", with_attachments=False, with_metadata=False)

        self.assert_calls()

        assert all(map(lambda x: "attachment" not in x and "metadata" not in x, ls))
        assert len(ls) == 4

    def test_download_filter_only_documents(self, monkeypatch, client):
        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        # type exist list
        self.__setup_mocks(monkeypatch)
        ls = dataset.download("test", with_attachments=False, with_metadata=False,
                              document_types=["xml"])

        self.assert_calls()

        assert all(map(lambda x: ".xml" in x, ls))

        # mix type exist/not exist list
        self.__setup_mocks(monkeypatch)
        ls = dataset.download("test", with_attachments=False, with_metadata=False,
                              document_types=["xml", "html"])

        self.assert_calls()

        assert all(map(lambda x: ".xml" in x, ls))

        # type not exist
        self.__setup_mocks(monkeypatch)
        ls = dataset.download("test", with_attachments=False, with_metadata=False,
                              document_types="html")

        self.assert_calls()

        assert len(ls) == 0

        # type exist list 2
        self.__setup_mocks(monkeypatch)
        ls = dataset.download("test", with_attachments=False, with_metadata=False,
                              document_types=["xml", "parquet"])

        self.assert_calls()

        assert all(map(lambda x: ".xml" in x or ".parquet" in x, ls))
        assert len(ls) == 4

    def test_download_skip_zero_document_folders(self,
                                                 monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        self.filter_mock.side_effect = iter([
            [
                self.key_latest_0,
                self.key_document1_attachment_0,
                # Removed the document from here!
                self.key_document1_metadata_0,
                self.key_document2_attachment_0,
                self.key_document2_document_0,
                self.key_document2_document_1,
                self.key_document2_metadata_0,
            ]
        ])

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        # there are only attachment and metadata in document_id=2
        ls = dataset.download(
            "test",
            skip_folders_with_zero_matching_documents=True
        )

        self.assert_calls()

        assert all(['document_id=2' in x for x in ls]), f'List: {ls}'
        assert len([x for x in ls if 'document_id=1' in x]) == 0, \
            f'Excepted 0 for document_id=1. List: {ls}'
        assert len([x for x in ls if 'document_id=2' in x]) == 4, \
            f'Excepted 4 for document_id=2. List: {ls}'

        self.__setup_mocks(monkeypatch)

        self.filter_mock.side_effect = iter([
            [
                self.key_latest_0,
                self.key_document1_attachment_0,
                # Removed the document from here!
                self.key_document1_metadata_0,
                self.key_document2_attachment_0,
                self.key_document2_document_0,
                self.key_document2_document_1,
                self.key_document2_metadata_0,
            ]
        ])

        ls = dataset.download(
            "test",
            skip_folders_with_zero_matching_documents=False
        )

        self.assert_calls()

        assert any(['document_id=2' in x for x in ls])
        assert len([x for x in ls if 'document_id=1' in x]) == 2, \
            'Excepted 2 for document_id=1'
        assert len([x for x in ls if 'document_id=2' in x]) == 4, \
            'Excepted 4 for document_id=2'

    def test_download_metadata(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        # there are only attachment and metadata in document_id=3
        ls = dataset.download_metadata("test")

        self.assert_calls()

        assert all(map(lambda x: "metadata" in x or '.latest' in x, ls)), \
            f'List: {ls}'

    def test_download_attachments(self, monkeypatch, client):
        self.__setup_mocks(monkeypatch)

        dataset = client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Unstructured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

        # there are only attachment and metadata in document_id=3
        ls = dataset.download_attachments("test")

        self.assert_calls()

        assert all(map(lambda x: "attachment" in x or '.latest' in x, ls)), \
            f'List: ls'


    #check against dl-5214



import json
import os
import time
from urllib.parse import urlencode

import keyring
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait
import pytest
import unittest
import jwt
from unittest.mock import MagicMock, patch, Mock
import httpretty

from dli import __version__, __product__
from dli.client.components.urls import identity_urls
from dli.client.dli_client import Session, DliClient
from dli.client.environments import _Environment
from dli.client.session import _start_session
from dli.client import session
from dli.client.exceptions import (
    DatalakeException,
    InsufficientPrivilegesException,
    UnAuthorisedAccessException
)
from dli.siren import PatchedSirenBuilder

environ = MagicMock(catalogue='http://catalogue.local',
                    accounts='')


def _get_token(payload, secret: str) -> str:
    token_encoded = jwt.encode(payload, secret)
    if isinstance(token_encoded, bytes):
        return token_encoded.decode('utf-8')
    else:
        return token_encoded


valid_token: str = _get_token({"exp": 9999999999}, 'secret')
expired_token: str = _get_token({"exp": 1111111111}, 'secret')
long_token: str = _get_token({"exp": 9999999999, "accounts": ["test"]*1000}, 'secret')


class SessionTestCase(unittest.TestCase):

    def test_can_decode_valid_jwt_token(self):
        ctx = Session(
            access_id=None,
            secret_key="key",
            environment=environ,
            host=None,
            auth_key=valid_token,
            logger=MagicMock()
        )

        self.assertFalse(ctx.has_expired)

    def test_can_detect_token_is_expired(self):
        ctx = Session(
            access_id=None,
            secret_key="key",
            environment=environ,
            host=None,
            auth_key=expired_token,
            logger=MagicMock()
        )
        self.assertTrue(ctx.has_expired)


class SessionRequestFactoryTestCase(unittest.TestCase):

    @pytest.fixture(autouse=True)
    def session(self):
        self.session = Session(
            access_id=None, secret_key=None, environment=environ,
            host=None,
            auth_key=valid_token,
            logger=MagicMock()
        )

    @httpretty.activate
    def test_response_403_raises_InsufficientPrivilegesException(self):
        response_text = 'Insufficient Privileges'
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/test',
            status=403, body=response_text
        )

        with self.assertRaises(InsufficientPrivilegesException):
            self.session.get('/test')

    @httpretty.activate
    def test_response_401_raises_UnAuthorisedAccessException(self):
        response_text = 'UnAuthorised Access'
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/test',
            status=401, body=response_text
        )


        with self.assertRaises(UnAuthorisedAccessException):
            self.session.get('/test')

    @httpretty.activate
    def test_response_500_raises_DatalakeException(self):
        response_text = 'Datalake server error'
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/test',
            status=500, body=response_text
        )

        with self.assertRaises(DatalakeException):
            self.session.get('/test')

    @httpretty.activate
    def test_sdk_version_is_included_in_header(self):
        httpretty.register_uri(
            httpretty.GET, 'http://catalogue.local/__api/',
            status=200, body="response"
        )
        # issue a request
        self.session.get('/__api/')

        request = httpretty.last_request()
        self.assertTrue("X-Data-Lake-SDK-Version" in request.headers)
        self.assertEqual(request.headers["X-Data-Lake-SDK-Version"], str(__version__))


class UnitTestSessionMock(Session):

    def __init__(self, id, pasw):
        self._get_SAM_auth_key = MagicMock()
        self._get_auth_key = MagicMock()
        self._get_decoded_token = MagicMock()
        self._get_expiration_date = MagicMock()
        self._set_mount_adapters = MagicMock()
        super().__init__(
            id, pasw, _Environment("http://catalogue.local"), "Test",
            logger=MagicMock(),
        )


class UnitTestClientMock(DliClient):

    def __init__(self, api_root, host=None, debug=None, strict=True,
                 access_id=None, secret_key=None, use_keyring=True):
        super().__init__("Test", access_id=access_id, secret_key=secret_key, use_keyring=use_keyring)

    def _new_session(self):
        return UnitTestSessionMock("Test", "Test")


class UnitTestSessionMockWithAuth(Session):

    def __init__(self, id, pasw):
        self.logger = MagicMock()
        self._set_mount_adapters = MagicMock()
        self.access_id = id
        self.secret_key = pasw
        self.auth_key = None
        self._environment = _Environment("http://catalogue.local")
        self.host = None
        self.use_keyring = True
        self.siren_builder = PatchedSirenBuilder()


class UnitTestClientMockWithAuth(UnitTestClientMock):

    def _new_session(self):
        return UnitTestSessionMockWithAuth(self.access_id, self.secret_key)


class UnitTestSessionMockNoPrompt(Session):

    # required to prevent the web popup that we want to
    # launch ourselves / use selenium
    def __init__(self, id, pasw, env):
        super().__init__(id, pasw, env, "Test",
                         auth_prompt=False, logger=MagicMock())


class UnitTestClientMockNoPrompt(DliClient):

    def _new_session(self):
        return UnitTestSessionMockNoPrompt(None, None, self._environment)


@pytest.fixture
def clear_keyring(monkeypatch):
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: None)
    monkeypatch.setattr(keyring, 'set_password', lambda _, app, val: None)

@pytest.fixture
def mock_session(clear_keyring):
    yield UnitTestSessionMock

@pytest.fixture
def mock_del_env_user(monkeypatch):
    monkeypatch.delenv("DLI_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("DLI_SECRET_ACCESS_KEY", raising=False)

@pytest.fixture
def real_client(request):
    # This is the S3 address used for the QA environment.
    api_root = os.environ[f'{request.param[0]}_API_URL']
    accessid = os.environ[f'{request.param[0]}_ACCESS_ID']
    secretkey = os.environ[f'{request.param[0]}_SECRET_ACCESS_KEY']
    client = None

    if request.param[1] == "Credentials":
        client = DliClient(
            api_root=api_root,
            access_id=accessid,
            secret_key=secretkey
        )
    elif request.param == 'API':
        pass
    else:
        pass

    yield client


@pytest.fixture
def mock_create_env_user(monkeypatch):
    monkeypatch.setenv("DLI_ACCESS_KEY_ID", "TestingUser")
    monkeypatch.setenv("DLI_SECRET_ACCESS_KEY", "TestingPass")


def test_credentials_no_api_key(monkeypatch, mock_create_env_user, recwarn):
    # test flow when u/p set and no api key set
    monkeypatch.setattr(DliClient, '_new_session', lambda x: x)
    dl = _start_session()
    assert(issubclass(type(dl), DliClient))
    assert(dl.secret_key is not None and dl.access_id is not None)


def test_credentials_session(mock_create_env_user, mock_session):
    # test flow when u/p set and no api key set

    sesh = mock_session(os.environ["DLI_ACCESS_KEY_ID"],
                        os.environ["DLI_SECRET_ACCESS_KEY"])
    assert sesh._get_SAM_auth_key.call_count == 1


def test_api_key_deprecation_warning(recwarn, monkeypatch):
    # test flow when u/p set and api key set
    # api key should override
    monkeypatch.setattr(DliClient, '_new_session', MagicMock())
    dl = _start_session(api_key="Test")
    assert(issubclass(type(dl), DliClient))
    assert "`api_key` will be deprecated in the future" \
           in ",".join([str(x.message) for x in recwarn.list])
    assert(dl.secret_key is not None and dl.access_id is None)


def test_credentials_and_api_key(
    clear_keyring, mock_create_env_user, monkeypatch
):
    # test flow when u/p set and api key set
    # api key should override
    monkeypatch.setattr(DliClient, '_new_session',  MagicMock())
    dl = _start_session(api_key="Test")
    assert (dl.secret_key is not None and dl.access_id is None)


def test_api_session(mock_create_env_user, mock_session):
    # test flow when u/p set and no api key set
    # (there will be no acces id set)
    sesh = mock_session(None, "Test")
    assert sesh._get_auth_key.call_count == 1


def test_credentials_and_no_api_key(
        clear_keyring, mock_create_env_user, monkeypatch, caplog
):
    # test flow when no credentials and api key set
    monkeypatch.setattr(session, "get_client", lambda: UnitTestClientMock)
    dl = _start_session()
    for x in caplog.records:
        assert("old" not in x.message)


def test_no_credentials_and_api_key(mock_del_env_user, monkeypatch, caplog):
    # test flow when no credentials and api key set
    monkeypatch.setattr(session, "get_client", lambda: UnitTestClientMock)
    dl = _start_session(api_key="Test")
    assert (dl.secret_key is not None and dl.access_id is None)
    for x in caplog.records:
        assert("new" not in x.message)


def test_api_key_auth(mock_del_env_user, monkeypatch):
    # test old auth process (_get_auth_key())
    api_response = valid_token
    monkeypatch.setattr(session, "get_client", lambda: UnitTestClientMockWithAuth)
    dl = _start_session(api_key="API")
    dl._session.post = MagicMock()
    dl._session.post.return_value.text = api_response

    dl._session._auth_init()
    assert(dl and dl._session.auth_key is not None)


def test_credentials_auth(clear_keyring, mock_create_env_user, monkeypatch):
    # test new auth process (_get_SAM_auth_key())
    sam_response = {
        "access_token": valid_token,
        "token_type": "Bearer", "expires_in": 3599}
    catalogue_response = {
        "access_token": valid_token,
        "token_type": "Bearer", "expires_in": 3599}

    monkeypatch.setattr(session, "get_client", lambda: UnitTestClientMockWithAuth)
    dl = _start_session()
    dl._session.post = MagicMock()
    dl._session.post.return_value.json.side_effect = [
        sam_response, catalogue_response
    ]

    dl._session._auth_init()
    assert(dl and dl._session.auth_key is not None)



def real_drive(port, postbox, env):
    chrome_options = Options()
    # chrome_options.add_argument("--headless")

    user = "dl_manual_test_2@ihsmarkit.com"
    pw = "-0o_p}i-Uz%*PQy_"

    driver = webdriver.Chrome(executable_path='tests/chromedriver',
                              options=chrome_options)

    driver.get(
        f"{env.catalogue}"
        f"{identity_urls.identity_postbox}"
        f"?postbox={postbox}"
    )

    WebDriverWait(driver, 120).until(
        ec.visibility_of_element_located((By.ID, "emailAddress"))
    ).send_keys(user)

    driver.find_element(By.ID, "continueButton").click()

    WebDriverWait(driver, 10).until(
        ec.visibility_of_element_located((By.ID, "i0116"))
    ).send_keys(user)

    driver.find_element(By.ID, "idSIButton9").click()
    WebDriverWait(driver, 10).until(
        ec.visibility_of_element_located((By.NAME, "passwd"))
    ).send_keys(pw)

    driver.find_element(By.ID, "idSIButton9").click()

    driver.find_element(By.ID, "idSIButton9").click()
    # need to add some time, else the client quits before we
    # complete localhost activity.
    time.sleep(2)
    driver.quit()


def test_reload_JWT_keyring(monkeypatch):
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: valid_token)

    dl = UnitTestClientMockWithAuth(api_root="Test")
    dl._session._get_web_auth_key = MagicMock()
    dl._session._auth_init()
    assert dl._session._get_web_auth_key.call_count == 0


def test_reload_long_JWT_keyring(monkeypatch):
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: long_token)

    dl = UnitTestClientMockWithAuth(api_root="Test")
    dl._session._get_web_auth_key = MagicMock()
    dl._session._auth_init()
    assert dl._session.auth_key == long_token
    assert dl._session._get_web_auth_key.call_count == 0


@patch("os.name","nt")
def test_store_long_long_keyring_for_windows(monkeypatch, caplog):

    def errorout(domain, app, pasw):
        raise OSError("Key is too long")

    monkeypatch.setattr(keyring, 'set_password', errorout)
    dl = UnitTestClientMockWithAuth(api_root="Test")

    def catch(port, postbox):
        dl._session.get = MagicMock()
        dl._session.get.return_value = Mock(status_code=200, text="test", content="test")

    dl._session._get_web_auth_key(callback=catch)
    # n.b we are gonna get this, but because we play with os.name
    # we will vomit another exception about trying to use a window cred mgr
    assert 'Key is too long' in caplog.text

@patch("os.name","nt")
def test_store_long_long_keyring(monkeypatch, caplog):

    class TestKeyring(keyring.backend.KeyringBackend):
        priority = 1
        dic = {}

        def set_password(self, servicename, username, password):
            TestKeyring.dic[servicename+username] = password

        def get_password(self, servicename, username):
            return TestKeyring.dic[servicename+username]

        def delete_password(self, servicename, username, password):
            del TestKeyring.dic[servicename+username]

    tk = TestKeyring()
    old_keyring = keyring.get_keyring()
    keyring.set_keyring(tk)
    old = keyring.set_password

    def error_then_do(domain, app, pasw):
        keyring.set_password = old
        # impersonate the first Windows error
        raise OSError("Key is too long")

    keyring.set_password = error_then_do

    dl = UnitTestClientMockWithAuth(api_root="Test")

    def catch(port, postbox):
        dl._session.get = MagicMock()
        dl._session.get.return_value = Mock(status_code=200, text=valid_token, content="test")

    dl._session._get_web_auth_key(callback=catch)
    # n.b we are gonna get this, but because we play with os.name
    # we will vomit another exception about trying to use a window cred mgr
    keyring.set_keyring(old_keyring)
    assert len(tk.dic.keys()) == 2

    assert tk.dic["ihsm-datalakehttp://catalogue.local"] == "**split**1"
    assert tk.dic["ihsm-datalakehttp://catalogue.local-0"] == valid_token


@patch("os.name","nt")
def test_store_long_long_keyring_split(monkeypatch, caplog):

    class TestKeyring(keyring.backend.KeyringBackend):
        priority = 1
        dic = {}

        def set_password(self, servicename, username, password):
            TestKeyring.dic[servicename+username] = password

        def get_password(self, servicename, username):
            return TestKeyring.dic[servicename+username]

        def delete_password(self, servicename, username, password):
            del TestKeyring.dic[servicename+username]

    tk = TestKeyring()
    old_keyring = keyring.get_keyring()
    keyring.set_keyring(tk)
    old = keyring.set_password

    def error_then_do(domain, app, pasw):
        keyring.set_password = old
        raise OSError("Key is too long")


    keyring.set_password = error_then_do

    dl = UnitTestClientMockWithAuth(api_root="Test")

    def catch(port, postbox):
        dl._session.get = MagicMock()
        dl._session.get.return_value = Mock(status_code=200, text=long_token, content="test")

    dl._session._get_web_auth_key(callback=catch)
    # n.b we are gonna get this, but because we play with os.name
    # we will vomit another exception about trying to use a window cred mgr
    keyring.set_keyring(old_keyring)
    assert len(tk.dic.keys()) == 11

    assert tk.dic["ihsm-datalakehttp://catalogue.local"] == "**split**10"
    for x in range(int(tk.dic["ihsm-datalakehttp://catalogue.local"].split("**split**")[-1])):
        assert tk.dic[f"ihsm-datalakehttp://catalogue.local-{x}"] == \
               long_token[x*1000:((1+x)*1000)]

def test_no_reload_delete_expired_JWT_keyring(monkeypatch):

    dl = UnitTestClientMockWithAuth(api_root="Test")
    # we definitely set it in the first place
    keyring.set_password(
        __product__, dl._session._environment.catalogue, valid_token)
    # later 'expires'
    orig = keyring.get_password
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: expired_token)

    dl._session._get_web_auth_key = MagicMock()
    dl._session._get_web_auth_key.return_value = valid_token
    dl._session._auth_init()
    assert dl._session._get_web_auth_key.call_count == 1
    assert(orig(__product__, dl._session._environment.catalogue) is None)


@pytest.mark.skipif(os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
def test_set_JWT_keyring_web_auth_key(monkeypatch):
    orig = keyring.get_password
    monkeypatch.setattr(keyring, 'get_password', lambda _, app: None)
    dl = UnitTestClientMockWithAuth(api_root="Test")

    def catch(port, postbox):
        dl._session.get = MagicMock()
        dl._session.get.return_value = Mock(status_code=200, text=valid_token, content="test")

    dl._session._get_web_auth_key(callback=catch)
    assert(orig(__product__, dl._session._environment.catalogue)
           == valid_token)


@pytest.mark.xfail
@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
def test_web_flow(clear_keyring, mock_del_env_user, monkeypatch):

    monkeypatch.setattr(session, "get_client", lambda: UnitTestClientMockNoPrompt)
    dl = _start_session(root_url="https://catalogue-dev.udpmarkit.net")

    # this is what get_web_auth_key is otherwise doing,
    # but we need to unblock to allow selenium to operate.
    # so we register a custom default callback to call selenium
    # rather than the web browser
    def cb(port, postbox):
        real_drive(port, postbox, dl._environment)

    with patch.object(Session._get_web_auth_key, '__defaults__', (cb,)):
        dl._session._auth_init()

    assert dl._session.auth_key is not None

@pytest.mark.xfail
@pytest.mark.skipif(os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
@pytest.mark.parametrize('real_client', [('QA', "Credentials")],
                         indirect=['real_client'])
def test_credentials_on_catalogue(clear_keyring, real_client):
    # test that new auth actually works and returns JWT

    assert(real_client.datasets.get(
        "autotestdatasetItsdeclarationRomanZimmermannandpeople"
    ) is not None)

@pytest.mark.xfail
@pytest.mark.skipif(True or os.environ.get('CI_PYPI_USER') is not None,
                    reason="Not runnable on gitlab - 8080 closed")
@pytest.mark.integration
@pytest.mark.parametrize('real_client', [('QA', "Credentials")],
                         indirect=['real_client'])
def test_webflow_on_catalogue(monkeypatch, real_client):

    monkeypatch.setattr(session, "get_client", lambda: UnitTestClientMockNoPrompt)
    dl = _start_session(root_url="https://catalogue-dev.udpmarkit.net")

    # this is what get_web_auth_key is otherwise doing,
    # but we need to unblock to allow selenium to operate.
    # so we register a custom default callback to call selenium
    # rather than the web browser
    def cb(port, postbox):
        real_drive(port, postbox, dl._environment)

    with patch.object(Session._get_web_auth_key, '__defaults__', (cb,)):
        dl._session._auth_init()

    # todo - doesnt work with these u/p sam logins
    assert dl.datasets.get("Sheep") is not None


def test_no_system_keyring_installed(monkeypatch):
    keyring.core.set_keyring(keyring.core.fail.Keyring())
    monkeypatch.setattr("dli.client.dli_client.Session._reload_or_web_flow.__defaults__", (True,))
    try:
        from keyring.errors import NoKeyringError
        with pytest.raises(NoKeyringError):
            dl = _start_session(root_url="https://catalogue-dev.udpmarkit.net")
    except ImportError:
        with pytest.raises(RuntimeError):
            dl = _start_session(root_url="https://catalogue-dev.udpmarkit.net")


def test_no_system_keyring_ack(monkeypatch, capsys):
    keyring.core.set_keyring(keyring.core.fail.Keyring())
    a = Mock(status_code=200, text=valid_token, content="test")

    def t(pb):
        return a


    monkeypatch.setattr("dli.client.dli_client.Session._get_web_auth_key.__defaults__",
                        ( (lambda port, pb: t(pb)), ))
    monkeypatch.setattr(
        "dli.client.dli_client.Session.get",  (lambda port, pb: a)
    )

    dl = _start_session(root_url="https://catalogue-dev.udpmarkit.net")


    captured = capsys.readouterr()
    assert "no keyring manager available" in captured.out


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import textwrap
import warnings
from collections import OrderedDict
from typing import Dict

from dli.client import utils
from dli.models.paginator import Paginator
from dli.client.components.urls import package_urls
from dli.models import AttributesDict
from dli.models.dataset_model import DatasetModel


class PackageModel(AttributesDict):
    """
    Represents a package pulled back from Me consumables
    """

    @classmethod
    def _raw_dict(cls, v1_response):
        return v1_response['properties']

    @classmethod
    def _from_v1_response_to_v2(cls, v1_response):
        response = cls._client.session.get(
            package_urls.v2_package_by_id.format(
                id=v1_response['properties']['packageId']
            )
        )
        json = response.json()
        return cls(json)

    @classmethod
    def _from_v2_response_unsheathed(cls, response_json, page_size=None):
        """

        :param response_json:
        :param page_size: Unused value, added to keep the same interface as
        datasetModel's _from_v2_response_unsheathed.
        :return:
        """

        return cls({
            'data': response_json
        })

    def __init__(self, json):
        # We have to declare the attribute `documentation` because it is
        # referenced in the code of this class. This means that there will
        # be an `documentation` attribute even when there is zero
        # `documentation` attribute in the Catalogue response JSON.
        self.documentation = None
        self.__shape = None

        # Maintain compatibility between V1 (which had package_id) and V2
        # (which only has ID).
        if 'package_id' not in json['data']['attributes']:
            json['data']['attributes']['package_id'] = json['data']['id']

        super().__init__(id=json['data']['id'], **json['data']['attributes'])
        self._paginator = Paginator(
            package_urls.v2_package_datasets.format(id=self.id),
            self._client._DatasetFactory,
            self._client._DatasetFactory._from_v2_response_unsheathed
        )

    @property
    def shape(self):
        """
        :returns: Count the number of datasets.
        :rtype: Int
        """
        if self.__shape is None:
            self.__shape = len(self.datasets())
        return self.__shape

    def datasets(self) -> Dict[str, DatasetModel]:
        """
        :returns: Dictionary of datasets in a package.
        :rtype: OrderedDict[id: str, DatasetModel]
        """
        return OrderedDict([
            (v.short_code, v) for v in self._paginator
        ])

    def contents(self):
        """Print information about all the datasets in this package."""
        for p in self.datasets().items():
            print(str(p[1]))

    def metadata(self):
        """
        Once you have selected a package, you can print the metadata
        (the available fields and values).

        :example:

            .. code-block:: python
                # Get all packages.
                >>> packages = client.packages('example package')
                # Get metadata of the 'example package' package.
                >>> packages['example package'].metadata()

        :example:

            .. code-block:: python
                # Get all packages with a name containing the text 'example package'. Filtering
                # is done on the server side.
                >>> packages = client.packages('example package')
                # Get metadata of the 'example package' package.
                >>> packages['example package'].metadata()

        :return: Prints the metadata.
        """
        utils.print_model_metadata(self)

    def __str__(self):
        separator = "-"*80
        split_description = "\n".join(textwrap.wrap(self.description, 80))
        split_keywords = "\n".join(self.keywords or [])

        # documentation is not guaranteed to be available
        split_documentation = 'No documentation available.'
        if self.documentation is not None:
            split_documentation = "\n".join(textwrap.wrap(self.documentation, 80))

        # When the fields are re-named in the JSON, it breaks our mapping. Use
        #   return str(self.__dict__)
        # to get the new field names then update the below.
        return f"\nPACKAGE \"{self.name}\" " \
               f"(Contains: {self.shape} datasets)\n" \
               f">> ID: {self.id} \n" \
               f">> Accessible: {self.has_access}\n" \
               f"\n" \
               f"{split_description}\n" \
               f"Documentation: {split_documentation}\n\n" \
               f"Keywords:\n{split_keywords}\n" \
               f"{separator}"

    def __repr__(self):
        return f'<Package name={self.name}>'



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import os
import uuid
from abc import ABCMeta
from typing import List, Optional, Callable, Tuple, Iterator

from tqdm import tqdm

import boto3.s3.transfer

from dli.aws import create_refreshing_session, trace_logger, \
    is_hidden_file, get_s3_list, get_s3_list_filter_out_hidden, \
    get_most_recent_common_prefix
from dli.client import utils
from dli.models import AttributesDict
from dli.models.file_model import get_or_create_os_path


class DatasetModel(AttributesDict, metaclass=ABCMeta):

    def __init__(self, **kwargs):
        # We have to declare the attribute `load_type` because it is
        # referenced in the code of this class. This means that there will
        # be an `documentation` attribute even when there is zero
        # `documentation` attribute in the Catalogue response JSON.
        self.load_type: Optional[str] = None

        super().__init__(**kwargs, )

    def __endpoint_url(self):
        return f'https://{self._client._environment.s3_proxy}'

    def _check_access(self):
        if not self.has_access:
            # Link to the package page in the Catalogue UI where the user
            # can request access.
            # eg. https://catalogue.datalake.ihsmarkit.com/package/
            # 7eddd22c-2b3c-11eb-9dd7-f64fac425b13
            package_url = f'{self._client._environment.catalogue}/' \
                          f'package/{self.package_id}'
            raise Exception(
                'You require access to this package to retrieve its content. '
                f'To request access, please visit {package_url}'
            )

    @staticmethod
    def _get_transfer_manager(s3_client, config):
        return boto3.s3.transfer.create_transfer_manager(
            client=s3_client,
            config=config,
        )

    @property
    def id(self):
        return self.dataset_id

    def __list(
        self,
        request_id=None,
        custom_filter_func: Callable[[List[Tuple[str, int]]], List[Tuple[str, int]]] = None,
        filter_path: Optional[str] = None,
        absolute_path: bool = True,
        skip_hidden_files: bool = True,
    ) -> List[str]:
        """
        List all the paths to files in the dataset. Calls go via the
        Datalake's S3 proxy, so the returned paths will be returned in the
        style of the S3 proxy, so in the pattern
        `s3://organisation_short_code/dataset_short_code/<path>`.
        S3 proxy does not return direct paths to the real location on S3 as
        they may be sensitive.

        The list will filter out paths where:
        * The size is zero bytes.
        * The name of any partition within the path is prefixed with a `.` or
        a `_` as that means it is intended as a hidden file.
        * The name of any partition within the path is exactly the word
        `metadata`.
        * It is a directory rather than a file.
        * Files with the wrong extension for the dataset's type i.e for
        a .parquet dataset we will return only .parquet files, zero .csv
        files.

        This list will filter based on `load_type` of the dataset. When the
        dataset's `load_type` is `Incremental Load` then files will be listed
        from all of the `as_of_date` partitions on S3. When the
        dataset's `load_type` is `Full Load` then files will be listed only
        the most recent `as_of_date` partition on S3.
        Please see the support library documentation for more information
        about how the `load type` affects the data you can access in a dataset:
        https://supportlibrary.ihsmarkit.com/display/DLSL/Exploring+Datasets#ExploringDatasets-Howthe%60loadtype%60affectsthedatayoucanaccessinadataset

        Parameters
        ----------
        :param str request_id: Optional. Automatically generated value
            used by our logs to correlate a user journey across several
            requests and across multiple services.

        :param custom_filter_func: A function which may optionally filter the s3 paths returned during the first
        stage of download (namely path lookup as part of list is filtered).

        :param str filter_path: Optional. If provided only a subpath matching
            the filter_path will be matched. This is less flexible than using
            the `partitions` parameter, so we recommend you pass in
            `partitions` instead. The `partitions` can deal with the
            partitions in the path being in any order, but filter_path relies
            on a fixed order of partitions that the user needs to know ahead
            of time.

            Example usage to get all paths that start with the `as_of_date`
            2020. Note this is a string comparison, not a datetime comparison.

            .. code-block:: python

                dataset.list(filter_path='as_of_date=2020')

        :param bool absolute_path: True (default) for returning an
            absolute path to the path on the S3 proxy. False to return a
            relative path (this is useful if using a TransferManager).

        :param bool skip_hidden_files: True (default) skips files that have
            been uploaded to S3 by Spark jobs. These usually start with a
            `.` or `_`.
        """

        files = self.__list_with_size(
            request_id=request_id,
            custom_filter_func=custom_filter_func,
            filter_path=filter_path,
            absolute_path=absolute_path,
            skip_hidden_files=skip_hidden_files
        )
        # this could be a completely empty prefix.
        if files:
            paths, _ = zip(*files)
        else:
            return []

        return list(paths)

    def __list_with_size(
        self,
        request_id=None,
        custom_filter_func: Callable[[List[Tuple[str, int]]], List[Tuple[str, int]]] = None,
        filter_path: Optional[str] = None,
        absolute_path: bool = True,
        skip_hidden_files: bool = True,
    ) -> List[Tuple[str, int]]:
        """
        List all the paths to files in the dataset. Calls go via the
        Datalake's S3 proxy, so the returned paths will be returned in the
        style of the S3 proxy, so in the pattern
        `s3://organisation_short_code/dataset_short_code/<path>`.
        S3 proxy does not return direct paths to the real location on S3 as
        they may be sensitive.

        The list will filter out paths where:
        * The size is zero bytes.
        * The name of any partition within the path is prefixed with a `.` or
        a `_` as that means it is intended as a hidden file.
        * The name of any partition within the path is exactly the word
        `metadata`.
        * It is a directory rather than a file.
        * Files with the wrong extension for the dataset's type i.e for
        a .parquet dataset we will return only .parquet files, zero .csv
        files.

        This list will filter based on `load_type` of the dataset. When the
        dataset's `load_type` is `Incremental Load` then files will be listed
        from all of the `as_of_date` partitions on S3. When the
        dataset's `load_type` is `Full Load` then files will be listed only
        the most recent `as_of_date` partition on S3.
        Please see the support library documentation for more information
        about how the `load type` affects the data you can access in a dataset:
        https://supportlibrary.ihsmarkit.com/display/DLSL/Exploring+Datasets#ExploringDatasets-Howthe%60loadtype%60affectsthedatayoucanaccessinadataset

        Parameters
        ----------
        :param str request_id: Optional. Automatically generated value
            used by our logs to correlate a user journey across several
            requests and across multiple services.

        :param custom_filter_func: A function which may optionally filter the
        s3 paths returned during the first stage of download (namely path
        lookup as part of list is filtered).

        :param str filter_path: Optional. If provided only a subpath matching
            the filter_path will be matched. This is less flexible than using
            the `partitions` parameter, so we recommend you pass in
            `partitions` instead. The `partitions` can deal with the
            partitions in the path being in any order, but filter_path relies
            on a fixed order of partitions that the user needs to know ahead
            of time.

            Example usage to get all paths that start with the `as_of_date`
            2020. Note this is a string comparison, not a datetime comparison.

            .. code-block:: python

                dataset.list(filter_path='as_of_date=2020')

        :param bool absolute_path: True (default) for returning an
            absolute path to the path on the S3 proxy. False to return a
            relative path (this is useful if using a TransferManager).

        :param bool skip_hidden_files: True (default) skips files that have
            been uploaded to S3 by Spark jobs. These usually start with a
            `.` or `_`.
        """

        self._check_access()

        if request_id is None:
            request_id = str(uuid.uuid4())

        trace_logger.info(
            f"The following GET Requests to '{self.__endpoint_url()}' "
            f"will be using request_id: '{request_id}'"
        )

        def add_request_id_to_session(**kwargs):
            kwargs["request"].headers['X-Request-ID'] = request_id

        session = create_refreshing_session(
            dli_client_session=self._client.session,
            event_hooks=add_request_id_to_session
        )
        s3_resource = session.resource(
            's3',
            endpoint_url=self.__endpoint_url()
        )

        bucket = s3_resource.Bucket(
            self.organisation_short_code
        )

        filter_prefix = self.short_code + (
            '' if self.short_code.endswith('/') else '/'
        )

        if filter_path:
            filter_prefix = filter_prefix + filter_path.lstrip('/')

        if filter_path is None and skip_hidden_files:
            paths: List[Tuple[str, int]] = list(
                get_s3_list_filter_out_hidden(
                    resource=s3_resource,
                    bucket=bucket,
                    prefix=filter_prefix,
                    content_type=self.content_type,
                    load_type=self.load_type,
                    absolute_path=absolute_path,
                    organisation_short_code=self.organisation_short_code,
                )
            )
        else:
            # The user provided an explicit filter_path so give the
            # user the S3 list under that path. There is no need to do common
            # prefix filtering as the filter_path will point us to a folder
            # not in the top-level of the dataset (where the .hidden
            # directories usually appear).
            paths: List[Tuple[str, int]] = list(get_s3_list(
                bucket=bucket,
                prefix=filter_prefix,
                absolute_path=absolute_path,
                organisation_short_code=self.organisation_short_code,
            ))

        trace_logger.info(
            "Number of paths on S3 for this dataset: "
            f"'{len(paths)}'"
        )

        if skip_hidden_files:
            # Although we have skipped hidden files in the top-level, there
            # could be files under the prefix e.g. sometimes there are Spark
            # _SUCCESS files mixed in with the data at the leaf level of the
            # S3 tree.
            paths: List[Tuple[str, int]] = [
                path for path in paths
                if not any(is_hidden_file(split, self.content_type)
                           for split in path[0].split('/'))
            ]

            trace_logger.info(
                f"Number of paths on S3 for this dataset "
                f"after filtering out hidden files: '{len(paths)}'"
            )

        if custom_filter_func:
            len_before = len(paths)
            trace_logger.info(
                f'Filtering {len_before} paths according to this dataset '
                f"type's custom filter logic."
            )

            paths = custom_filter_func(paths=paths)
            trace_logger.info(
                f'Filtered out {len_before - len(paths)} / {len_before} paths '
                f'according to constraints.'
            )

            if len(paths) == 0:
                self._client.logger.warning(
                    f'There are no items matching your filter settings. Check '
                    f'your filter settings.\n'
                )
                return []

        return paths

    def __download(
        self,
        destination_path: str,
        flatten: Optional[bool] = False,
        filter_path: Optional[str] = None,
        custom_filter_func: Callable[[List[str]], List[str]] = None,
        skip_hidden_files: Optional[bool] = True
    ) -> List[str]:
        """
        Equivalent to the exposed `download` method. Used as a base for
        all models

        Additional Parameters
        ----------
        :param bool flatten: The flatten parameter default behaviour
            (flatten=False) will allow the user to specify that they would
            like to keep the underlying folder structure when writing the
            downloaded files to disk.
            When “flatten” is set to True the
            folder structure and download all data to the specified path in
            and retains only the original file name. This is not useful if the
            file names are identical for multiple files.

        :param custom_filter_func: A function which may optionally filter
        the s3 paths returned during the first stage of download (namely
        path lookup as part of list is filtered).
        """

        self._check_access()
        request_id = str(uuid.uuid4())

        files = self.__list_with_size(
            request_id=request_id,
            custom_filter_func=custom_filter_func,
            filter_path=filter_path,
            absolute_path=False,  # TransferManager uses relative paths.
            skip_hidden_files=skip_hidden_files,
        )

        if files:
            paths, sizes = zip(*files)
        else:
            return []

        def __add_request_id_to_session(**kwargs):
            kwargs["request"].headers['X-Request-ID'] = request_id

        s3_resource = create_refreshing_session(
            dli_client_session=self._client.session,
            event_hooks=__add_request_id_to_session
        ).resource(
            's3',
            endpoint_url=self.__endpoint_url()
        )

        s3_client = s3_resource.meta.client

        # multipart_threshold -- The transfer size threshold for which
        # multipart uploads, downloads, and copies will automatically
        # be triggered.
        # 500 GB should be a large enough number so that the download never
        # triggers multipart download on cloudfront.
        KB = 1024
        GB = KB * KB * KB
        multipart_threshold = 500 * GB

        # Use an environment variable to disable the entire progressbar
        # wrapper. This can be necessary for monitors where the bar will not
        # be correctly formatted in DataDog.
        disable_progress_bar = \
            True if os.environ.get('DLI_DISABLE_PROGRESS_BAR', '').lower() \
            == 'true' else False

        progress = tqdm(
            desc='Downloading ',
            total=sum(sizes), unit='B', unit_scale=1,
            position=0,
            disable=disable_progress_bar,
        )

        config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=multipart_threshold,
        )

        transfer_manager_context = DatasetModel._get_transfer_manager(
            s3_client,
            config
        )

        warns = []
        with transfer_manager_context as transfer_manager:
            with progress as pbar:
                # Create progress_callback outside of the loop.
                progress_callback = boto3.s3.transfer.ProgressCallbackInvoker(
                    pbar.update
                )
                subscribers = [progress_callback]

                _paths_and_futures = []

                for path in paths:

                    if not path.endswith('/'):
                        to_path = get_or_create_os_path(
                            s3_path=path,
                            to=destination_path,
                            flatten=flatten
                        )

                        trace_logger.debug(
                            f'Downloading {path} to: {to_path}...'
                        )

                        if os.path.exists(to_path):
                            warns.append(to_path)

                        # returns a future
                        future = transfer_manager.download(
                            self.organisation_short_code,
                            path,
                            to_path,
                            subscribers=subscribers,
                        )

                        _paths_and_futures.append((to_path, future))

                _successful_paths = []
                for path, future in _paths_and_futures:
                    try:
                        # This will block for this future to complete,
                        # but other futures will keep running in the
                        # background.
                        future.result()
                        _successful_paths.append(path)
                    except Exception as e:
                        message = f'Problem while downloading:' \
                                  f'\nfile path: {path}' \
                                  f'\nError message: {e}\n\n'

                        self._client.logger.error(message)
                        print(message)

            return _successful_paths

    def metadata(self):
        """
        Once you have selected a dataset, you can print the metadata (the
        available fields and values).

        :example:

            .. code-block:: python
                # Get all datasets.
                >>> datasets = client.datasets()

                # Get metadata of the 'ExampleDatasetShortCode' dataset.
                >>> datasets['ExampleDatasetShortCode'].metadata()

        :example:

            .. code-block:: python
                # Get an exact dataset using the dataset_short_code and
                # organisation_short_code.
                >>> dataset = client.get_dataset(dataset_short_code='ExampleDatasetShortCode', organisation_short_code='IHSMarkit')
                # Get metadata of the dataset.
                >>> dataset.metadata()

        :example:

            .. code-block:: python
                # Get all datasets.
                >>> dataset = client.datasets()['ExampleDatasetShortCode']
                # Get metadata of the dataset.
                >>> dataset.metadata()

        :return: Prints the metadata.
        """
        utils.print_model_metadata(self)

    def package(self):
        """
        Returns the parent package of this dataset based on the package_id stamped onto the dataset.

        # note there is no typing on this method to avoid a circular referencing.
        """
        return list(self._client.packages(search_term=f"package_id={self.package_id}").values())[0]


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import re
import warnings
from collections import OrderedDict
from typing import Dict, Optional, Union, List

import requests

from dli.client.exceptions import CatalogueEntityNotFoundException
from dli.models.paginator import Paginator
from dli.client.components.urls import package_urls
from dli.models.package_model import PackageModel


class PackageModule:
    """
    This class is bound to the client and can not be initalised
    directly.

    It is accessed by:

    .. code:: python

        client.packages


    Documentation on ``client.packages()`` is refered to
    as ``__call__``
    """

    @staticmethod
    def _search_terms_to_filters(search_terms, start_count):

        def to_dict(k, op, v):
            return {'key': k.strip(), 'op': op.strip(), 'val': v.strip()}

        def partition_search_term(x):
            split = re.split("(<|>|=>|<=|=|!=| contains | like | ilike )", x)
            if len(split) == 3:
                return to_dict(*split)
            else:
                return None

        arguments = map(partition_search_term, search_terms)
        enumerated_partitions = [
            (x, y) for x, y in enumerate(arguments, start=start_count)
            if y is not None
        ]
        opmp = {
            "=": "eq",
            ">=": "gte",
            "<=": "lte",
            ">": "gt",
            "<": "lt",
        }

        to_numbered_filter = lambda partition: (lambda idx, x: {
            f'filter[{idx}][field]': x['key'],
            f'filter[{idx}][operator]': opmp.get(x['op'], x['op']),
            f'filter[{idx}][value]': x['val']

        })(*partition)

        yield from map(to_numbered_filter, enumerated_partitions)

    @staticmethod
    def _filter_creation(search_term, only_mine):
        bump = 0
        if only_mine:
            yield {
                'filter[0][value]': True,
                'filter[0][operator]': 'eq',
                'filter[0][field]': 'has_access',
            }
            bump = 1
        else:
            yield {}

        if type(search_term) is str:
            search_term = [search_term]

        if search_term and type(search_term) is list:
            yield from PackageModule._search_terms_to_filters(search_term, bump)

    def __call__(
        self, search_term: Optional[Union[str, List[str]]]=None,
        only_mine: Optional[bool]=False
    ) -> 'OrderedDict[str, PackageModel]':
        """ 

        See packages we can access at the top level.
        
        :Example:

            Get all packages:

            .. code:: python

                client.packages()

        :Example:

            Get **only** packages you have access to and
            can read the datasets of:

            .. code:: python

                client.packages(only_mine=True)

        :Example:

            Search for packages using `ilike`. This will find fields that are
            similar to value. This is recommended over using `like` as it
            performs case insensitive matching.

            Some examples of how the `ilike` logic matches in the database
            when there is a dataset named `example`:

            'example' ILIKE 'example'    true

            'example' ILIKE 'ex%'     true

            'example' ILIKE 'c'      false

            Example usage in the DLI:

            .. code:: python

                # Exact match `example` (case sensitive).
                client.packages(search_term='name=example')

                # Match `example` (case insensitive).
                client.packages(search_term='name ilike example')

                # Match `example` (case insensitive) followed by any characters.
                client.packages(search_term='name ilike example%')

                # Match `example` (case insensitive) in the middle of any characters.
                client.packages(search_term='name ilike %example%')

        :Example:

            Search for packages with multiple search terms:

            .. code:: python

                client.packages(
                    search_term=[
                        'name=example',
                        'manager_id=example'
                    ]
                )

        :Example:

            Search for packages with name and only_mine:

            .. code:: python

                client.packages(
                    search_term='name=example',
                    only_mine=True,
                )


        :param search_term: search term for package. Can filter using a simple
            query language. E.G "field=value", "field>value", "field ilike value".
            Can pass in multiple values using a list of strings.

        :param only_mine: Specify whether to collect packages only
            accessible to you (user) or to discover packages that you may
            want to discover. Set to True searches and returns only those
            packages which you also have access to and also match the
            search_term conditions. Set to false (default) searches and
            returns all packages matching the search_term conditions.


        .. seealso::

            :ref:`Package Attributes <package-attributes>`
                Definition of a Package, and what fields you can filter on (all fields labeled as "dynamic".

        """

        # search_term = ["y=z", "a>c", "d", "f contains pull", "f ilike pig"]
        filters = {}
        for x in PackageModule._filter_creation(search_term, only_mine):
            filters.update(x)

        search_paginator = Paginator(
            package_urls.v2_package_index,
            self._client._Package,
            self._client._Package._from_v2_response_unsheathed,
            page_size=5000,
            filters=filters
        )

        return OrderedDict([(v.name, v) for v in search_paginator])

    def get(self, name: str) -> PackageModel:
        """
        Find a PackageModel with the matching name. If not found then
        returns None.

        :param name: The name of the package to
            collect (short codes are yet to be implemented for packages)

        :Example:

            Search for packages with multiple search terms:

            .. code:: python

                # NOTE: this will break in future when package
                # short codes supersede names. Use at your own
                # risk.
                client.package.get('package name')

        """

        # NOTA BENE - this is commented out until such a time that package shortcodes are being implemented
        # THE BELOW MESSAGE SHOULD THEN BE SHOWN UNTIL SUCH A TIME THAT EVERYONE HAS UPGRADED TO AT LEAST
        # THE FIRST RELEASE VERSION WHICH AGAIN SHOWS IT.
        # warnings.warn(
        #     'Getting a package by name'
        #     'will be deprecated in future. Short-codes will replace this.',
        #     PendingDeprecationWarning
        # )

        res = self._client.packages(search_term=[f"name={name}"]).get(name)
        if res:
            return res
        else:
            if self._client.strict:
                r = requests.Response()
                r.status_code = 404
                r.request = (lambda: True)
                r.request.method = f"No such package {name}"
                r.request.url = ""
                raise CatalogueEntityNotFoundException(
                    message=f"No such package {name}",
                    response=r
                )
            else:
                return None


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import itertools
import logging
import re
from functools import partial
from typing import Optional, List, Union, Tuple

from dli.client.aspects import analytics_decorator, logging_decorator
from dli.models import log_public_functions_calls_using
from dli.models.dataset_model import DatasetModel

trace_logger = logging.getLogger('trace_logger')


class UnstructuredDatasetModel(DatasetModel):

    def __init__(self, **kwargs):

        if "warn" in kwargs:
            warn = kwargs.pop('warn')
            # if warn:
            #     self._client.logger.warning(
            #         "WARN: This is an Unstructured dataset.\n"
            #         "Available methods and method arguments may differ, "
            #         "please see the documentation."
            #     )
        super().__init__(**kwargs, )

    class __DeferInCaseCall(object):
        blurb = """
        Unstructured datasets are different to structured datasets. They have 
        different methods, method arguments and properties.
        You can tell if it is an Unstructured dataset using `ds.content_type` 
        or by `print(ds)`.
        Please see our documentation on using Structured or Unstructured 
        dataset type respectively.
        """

        def __init__(self, name, force_call=False):
            self.name = name
            self.force_call = force_call

        def __repr__(self):
            if self.force_call:
                return self.__call__()
            else:
                raise AttributeError(
                    "This is an Unstructured dataset. There is no such "
                    f"property `{self.name}`.\n{self.blurb}"
                )

        def __call__(self, *args, **kwargs):

            def convert_list(args):
                def formatter(x):
                    if len(x) > 10:
                        return "'" + x[:10] + "...'"
                    else:
                        return "'" + x + "'"

                def joiner(y):
                    return ",".join(list(map(
                        lambda x: x[0] + "=" + formatter(str(x[1])), y.items()
                    )))

                def stringify(x):
                    if type(x) is str:
                        return x
                    elif type(x) is tuple:
                        return ",".join(x)
                    elif type(x) is dict:
                        return joiner(x)
                    else:
                        return str(x)

                return ",".join([stringify(x) for x in args if x])

            raise AttributeError(
                "This is an Unstructured dataset. There is no such method "
                f"`{self.name}({convert_list(args)})`.\n" +
                (f"Method `{self.name}` EXISTS HOWEVER, SO YOU SEEM TO HAVE "
                 "SPECIFIED SOME BAD ARGUMENTS!"
                 if hasattr(UnstructuredDatasetModel, self.name) else "") +
                f"\n{self.blurb}"
            )

    def __getattr__(self, name):

        def wrapper(*args, **kwargs):
            return UnstructuredDatasetModel.__DeferInCaseCall(name)

        return wrapper()

    def __repr__(self):
        return f'<Unstructured Dataset short_code={self.short_code}>'

    def __str__(self):
        return f"UNSTRUCTURED DATASET \"{self.short_code}\""

    def _type_exception_handler(self, exception, func, args, kwargs):
        """
        Needed by aspects.py in order to throw a custom message if the
        function cant be called from it.
        Say, because the arguments are wrong, or the function doesnt exist.
        >N.B. COMMENT THIS OUT IF DEBUGGING!
        """
        if "unexpected" in str(exception):
            UnstructuredDatasetModel.__DeferInCaseCall(func.__name__)(args,
                                                                      kwargs)
        else:
            raise Exception(str(exception))

    def __custom_filter(
        self,
        with_attachments: bool,
        with_metadata: bool,
        with_documents: bool,
        document_types: Optional[Union[str, List[str]]],
        skip_folders_with_zero_matching_documents: bool,
        paths: List[Tuple[str, int]]
    ) -> List[Tuple[str, int]]:
        re_attachments = re.compile("^[^/]+\/[^/]+\/attachment\/.*")
        re_metadata = re.compile("^[^/]+\/[^/]+\/metadata\/.*")
        re_documents = re.compile("^[^/]+\/[^/]+\/document\/.*")

        get_path = lambda x: x[0]

        if not with_attachments:
            paths = filter(
                lambda x: not re_attachments.match(get_path(x)), paths
            )
            trace_logger.info(
                f'Filtering out attachments'
            )

        if not with_metadata:
            paths = filter(
                lambda x: not re_metadata.match(get_path(x)), paths
            )
            trace_logger.info(
                f'Filtering out metadata'
            )

        if not with_documents:
            paths = filter(
                lambda x: not re_documents.match(get_path(x)), paths
            )
            trace_logger.info(
                f'Filtering out documents'
            )
        else:
            paths = list(paths)
            docs = list(filter(
                lambda x: re_documents.match(get_path(x)), paths
            ))
            if len(docs) == 0 and skip_folders_with_zero_matching_documents:
                self._client.logger.warning(
                    'There are no documents in the dataset, so it will not '
                    'be downloaded.'
                    f'\nThere may be metadata or attachments however, repeat '
                    f'the call with '
                    f'`skip_folders_with_zero_matching_documents`=False\n'
                    f'if these are still required.'
                )
                return []

            if document_types:
                paths = filter(
                    lambda x: not re_documents.match(get_path(x)), paths
                )
                if type(document_types) is str:
                    d = [document_types]
                else:
                    d = document_types

                dtypes = '|'.join(d)
                dtypes = dtypes.lower()
                self._client.logger.info(
                    f'Filtering out documents not of type(s): {dtypes}'
                )

                re_filetype = re.compile(f"^(.+/)*(.+).({dtypes})$")
                matching_docs = filter(
                    lambda x: re_filetype.match(get_path(x).lower()), docs)

                paths = list(itertools.chain(paths, matching_docs))

            # TODO need another bit in here to cluster by surviving document_id
            # which still have documents
            if skip_folders_with_zero_matching_documents:
                re_group = re.compile("^[^/]+\/([^/]+)\/document\/.*")
                survivors = []
                doc_ids = "|".join(set(x[0] for x in map(
                    lambda x: re_group.findall(get_path(x)), paths) if x))
                re_survivors = re.compile(f"^[^/]+/({doc_ids})/[^/]*/.*")
                paths = list(
                    filter(lambda x: re_survivors.match(get_path(x)), paths))

        return list(paths)

    def __common_download(self, destination_path, **kwargs):
        return self._DatasetModel__download(
            destination_path=destination_path,
            flatten=False,
            custom_filter_func=partial(
                self.__custom_filter, **kwargs
            ),
            # we dont know if there are hidden yet.
            # may need to set to True
            skip_hidden_files=False
        )

    def list(
        self,
        request_id=None,
        absolute_path: bool = True,
        skip_hidden_files: bool = False,
    ) -> List[str]:
        """
        List all the paths to files in the dataset. Calls go via the
        Datalake's S3 proxy, so the returned paths will be returned
        in the
        style of the S3 proxy, so in the pattern
        `s3://organisation_short_code/dataset_short_code/<path>`.
        S3 proxy does not return direct paths to the real location
        on S3 as
        they may be sensitive.

        The list will filter out paths where:
        * The size is zero bytes.
        * The name of any partition within the path is prefixed with
        a `.` or
        a `_` as that means it is intended as a hidden file.
        * The name of any partition within the path is exactly the word
        `metadata`.
        * It is a directory rather than a file.
        * Files with the wrong extension for the dataset's type i.e for
        a .parquet dataset we will return only .parquet files,
        zero .csv
        files.

        Parameters
        ----------
        :param str request_id: Optional. Automatically generated value
            used by our logs to correlate a user journey across several
            requests and across multiple services.

        :param bool absolute_path: True (default) for returning an
            absolute path to the path on the S3 proxy. False to
            return a
            relative path (this is useful if using a TransferManager).

        :param bool skip_hidden_files: True (default) skips files
        that have
            been uploaded to S3 by Spark jobs. These usually start
            with a
            `.` or `_`.
        """
        return self._DatasetModel__list(
            request_id=request_id,
            custom_filter_func=None,
            absolute_path=absolute_path,
            skip_hidden_files=skip_hidden_files
            # n.b. We need to decide if you can list by filtering
            # with_attachments, with_metadata, doc_types too
        )

    def download(
        self,
        destination_path: str,
        with_attachments: bool = True,
        with_metadata: bool = True,
        document_types: Optional[Union[str, List[str]]] = None,
        skip_folders_with_zero_matching_documents: bool = True
    ) -> List[str]:
        """
        Downloads the original unstructured dataset files from the data lake
        to local copy, to a destination of your choice.

        If the dataset does not have any document_id's within it a message
        will be shown stating this.

        If the filters do match any of the document_id's in the dataset such
        that no document_id will be downloaded
        (as skip_folders_with_zero_matching_documents=True), a message
        stating the filter does not match any documents with be shown
        stating this.

        For example:

        .. code-block:: python

            import dli
            client = dli.connect()
            dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
            # Download all the files in this dataset to the specified
            # directory.
            dataset.download('C:\\tmp') # Path for Windows temp directory.

        :example:
            /document_id=233/
                /attachments/      --> @see with_attachments
                   file.pdf
                   file2.xml
                /document/
                    file.html      --> downloaded
                    ...
                /metadata/         --> @see with_metadata
                    ...

        Parameters
        ----------
        :param str destination_path: required. The path on the system,
            where the files should be saved. Must be a directory, if does not
            exist, will be created.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                dataset.download('C:\\tmp') # Path for Windows temp directory.

            :example:

                /document_id=233/
                    /attachments/      --> downloaded only when "with_attachments" is set to True
                       file.pdf
                       file2.xml
                    /document/
                        file.html      --> downloaded
                        ...
                    /metadata/         --> @see with_metadata
                        ...

        :param bool with_attachments:
            An unstructured dataset comprises of one of more document folder
            prefixes e.g. {document_id=233, document_id=132}
            Specifies whether to download the attachments folder prefix within
            each document_id prefix. Default is True.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                # Download everything except the attachments directory.
                dataset.download(
                    'C:\\tmp',
                     with_attachments=False
                )

            :example:
                "with_attachments" is set to False

                /document_id=233/
                    /document/
                        file.html      --> downloaded
                        ...
                    /metadata/         --> @see with_metadata
                        ...

            :example:
                "with_attachments" is set to True

                /document_id=233/
                    /attachments/      --> downloaded only when "with_attachments" is set to True
                       file.pdf
                       file2.xml
                    /document/
                        file.html      --> downloaded
                        ...
                    /metadata/         --> @see with_metadata
                        ...

        :param bool with_metadata:
            An unstructured dataset comprises of one of more document folder
            prefixes e.g. {document_id=233, document_id=132}
            Specifies whether to download the metadata folder prefix within
            each document_id prefix. Default is True.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                # Download everything except the metadata directory.
                dataset.download(
                    'C:\\tmp',
                    with_metadata=False
                )

            :example:
                "with_metadata" is set to False

                /document_id=233/
                    /attachments/      --> @see with_attachments
                       ...
                    /document/
                        file.html      --> downloaded
                        ...

            :example:
                "with_metadata" is set to True

                /document_id=233/
                    /attachments/      --> @see with_attachments
                       ...
                    /document/
                        file.html      --> downloaded
                        ...
                    /metadata/         --> download only when "with_metadata" is set to True
                        summary.json

        :param str document_types: Optional[Union[str, List[str]]].
            A string or a list of strings representing the desired file types
            to download.
            Filter the documents in the /document/ folder prefix of each
            document_id in the dataset
            according to the file's prefix. Default is None.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                # Download everything that is a document with file type "html".
                # This will not filter inside the attachments or metadata
                # directories.
                document_types = ["html"]
                dataset.download(
                    'C:\\tmp',
                    document_types=document_types
                )

            :example:
                document_types is set to ["html"] or "html"

                /document_id=233/
                    /attachments/
                       file.pdf
                       ...
                    /document/
                        file.html      --> downloaded
                        file2.pdf      --> not downloaded
                        file3.xml      --> not downloaded
                        ...
                    /metadata/
                        summary.json
                        ...

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                # Download everything that is a document with file type "pdf"
                # or "xml".
                # This will not filter inside the attachments or metadata
                # directories.
                document_types = ["pdf", "xml"]
                dataset.download(
                    'C:\\tmp',
                    document_types=document_types
                )

            :example:
                document_types is set to ["pdf", "xml"]

                /document_id=233/
                    /attachments/
                       file.pdf
                       ...
                    /document/
                        file.html      --> not downloaded
                        file2.pdf      --> downloaded
                        file3.xml      --> downloaded
                        ...
                    /metadata/
                        summary.json
                        ...

        :param bool skip_folders_with_zero_matching_documents:
            After applying document_types the document_id's /document/ folder
            may be empty,
            since the file types may not match the chosen types.
            The document_id's without any post-filter documents are filtered
            out by default to minimize
            download and retain only the folders bearing document_id's that
            matched the filter.

            It may be wished to download the document_id regardless of the
            folder being empty if the metadata or
            attachments are needed despite an empty /document/ folder. This
            is achieved by setting the value to False.
            Default is True.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                # Download everything that is a document with file type "pdf"
                # or "xml".
                # This will not filter inside the attachments or metadata
                # directories.
                document_types = ["xml"]
                dataset.download(
                    'C:\\tmp',
                    document_types=document_types,
                    skip_folders_with_zero_matching_documents=True
                )

            :example:
                when document_types = ["xml"] and skip_folders_with_zero_matching_documents = True

                /document_id=233/     --> not downloaded as /document/ empty
                after filter
                    /attachments/
                       file.pdf
                       ...
                    /document/      --> not downloaded
                        file.html      --> not downloaded
                        file2.pdf      --> not downloaded
                        ...
                    /metadata/
                        summary.json
                        ...

            :example:
                when document_types = ["xml"] and skip_folders_with_zero_matching_documents = False

                /document_id=233/      --> downloaded despite /document/ being empty after filter
                    /attachments/
                       file.pdf
                    /metadata/
                        summary.json
        """

        return self.__common_download(
            destination_path,
            with_attachments=with_attachments,
            with_metadata=with_metadata,
            with_documents=True,
            document_types=document_types,
            skip_folders_with_zero_matching_documents=skip_folders_with_zero_matching_documents
        )

    def download_metadata(self, destination_path: str) -> List[str]:
        """
        Helper function to call `download` with the settings to only download
        files that are in the /metadata directory.

        :param str destination_path: required. The path on the system,
            where the files should be saved. Must be a directory, if does not
            exist, will be created.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                dataset.download('C:\\tmp') # Path for Windows temp directory.

            :example:

                /document_id=233/
                    /attachments/      --> downloaded only when "with_attachments" is set to True
                       file.pdf
                       file2.xml
                    /document/
                        file.html      --> downloaded
                        ...
                    /metadata/         --> @see with_metadata
                        ...
        :return: List[str]
        """

        return self.__common_download(
            destination_path=destination_path,
            with_attachments=False,
            with_metadata=True,
            with_documents=False,
            document_types=None,
            skip_folders_with_zero_matching_documents=False
        )

    def download_attachments(self, destination_path: str) -> List[str]:
        """
        Helper function to call `download` with the settings to only download
        files that are in the /attachments directory.

        :param str destination_path: required. The path on the system,
            where the files should be saved. Must be a directory, if does not
            exist, will be created.

            .. code-block:: python

                import dli
                client = dli.connect()
                dataset = client.datasets.get(<dataset shortcode>, Optional:<organisation_short_code>)
                dataset.download('C:\\tmp') # Path for Windows temp directory.

            :example:

                /document_id=233/
                    /attachments/      --> downloaded only when "with_attachments" is set to True
                       file.pdf
                       file2.xml
                    /document/
                        file.html      --> downloaded
                        ...
                    /metadata/         --> @see with_metadata
                        ...
        :return: List[str]
        """

        return self.__common_download(
            destination_path=destination_path,
            with_attachments=True,
            with_metadata=False,
            with_documents=False,
            document_types=None,
            skip_folders_with_zero_matching_documents=False
        )


log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['dataset_id']
)(UnstructuredDatasetModel)



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
from pathlib import Path

from dli.models import AttributesDict


def makedirs(path, exist_ok=False):
    Path(path).mkdir(parents=True, exist_ok=exist_ok)

def to_camel_case(snake_str):
    split = snake_str.split('_')
    camel_case_parts = [split[0].lower()] + [s.title() for s in split[1:]]
    return ''.join(camel_case_parts)

def to_camel_cased_dict(dictionary):
    return {to_camel_case(key): value for (key, value) in dictionary.items()}

def filter_out_unknown_keys(dictionary, keys_set):
    return {k: v for (k, v) in dictionary.items() if k in keys_set}

def ensure_count_is_valid(count):
    count = int(count)
    if count <= 0:
        raise ValueError("`count` should be a positive integer")


def print_model_metadata(model: AttributesDict):
    for (k, v) in vars(model).items():
        if k == "documentation" and type(v) is str:
            val = v[0:100].replace('\n', ' ').replace('\r', '') + "..." \
                if len(v) > 100 else v
        else:
            val = v

        if not (hasattr(v, '__dict__') or k.startswith("_")):
            print(f"{k}: " + f"{val}")



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#

class ModelDescriptor:
    """
    This class is responsible for extending the base type passed
    into the __init__ method with the instance of DliClient it has
    been created within, following the descriptor pattern.

    What this means practicably, is that under _client attribute of the 'new'
    type (class instance) there is a backreference to the DliClient,
    which then permits the type to access the shared session object of DliClient,
    rather than having to pass the session into each instance.

    Using an instance instantiated from the base type will not have the
    _client attribute available.
    """

    def __init__(self, model_cls=None):
        self.model_cls = model_cls

    def __get__(self, instance, owner):
        """Returns a model thats bound to the client instance"""
        return type(
            self.model_cls.__name__, (self.model_cls, ),
            {
                '_client': instance
            }
        )


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import warnings
import os


warnings.filterwarnings(
    'always', module="dli"
)

warnings.filterwarnings(
    'ignore', module="dli", category=ResourceWarning
)

warnings.filterwarnings(
    action='ignore',
    category=FutureWarning,
    message='^pandas.util.testing is deprecated'
)

__version__ = '1.13.0'
__product__ = "ihsm-datalake"


try:
    import simplejson as _  # noqa
    warnings.warn(
        'Incompatible Package `simplejson`.\n\n'
        '\t`simplejson` is a backport of the built in json library in Python. '
        'It contains subtle differences, and is not intended for use beyond '
        'Python 2.6. Please uninstall `simplejson` by running:\n\n'
        '\t\tpip uninstall simplejson\n\n'
        '\tOr run the DLI from a virtual environment as it is known to cause '
        'issues within the DLI.\n',
        ImportWarning
    )
except ImportError:
    pass


def connect(
    api_key=None,
    root_url="https://catalogue.datalake.ihsmarkit.com/__api",
    host=None,
    debug=None,
    strict=True,
    use_keyring=True,
    log_level=None,
):
    """
    Entry point for the Data Lake SDK, returns a client instance that
    can be used to consume or register datasets.

    Example for starting a session:


    .. code:: python

        import dli
        client = client.connect()

    :param str api_key: Your API key, can be retrieved from your dashboard in
                        the Catalogue UI.
    :param str root_url: Optional. The environment you want to point to. By default it
                        points to Production.
    :param str host: Optional. Advanced usage, meant to force a hostname when DNS resolution
                     is not reacheable from the user's network.
                     This is meant to be used in conjunction with an
                     IP address as the root url.
                     Example: catalogue.datalake.ihsmarkit.com

    :param bool debug: Optional. Log SDK operations to a file in the current working
                       directory with the format "sdk-{end of api key}-{timestamp}.log"

    :param bool strict: Optional. When True, all exception messages and stack
                        trace are printed. When False, a shorter message is
                        printed and `None` should be returned.
    :param bool use_keyring: Optional. When True, cache the JWT in the system keyring
                             and retrieve that JWT from the system keyring if set.
                             Otherwise disable the keyring completely.

    :returns: Data Lake interface client
    :rtype: dli.client.dli_client.DliClient

    .. deprecated:: 1.8.3

        `api_key`: please use the basic flow or client credentials.

    """
    from dli.client.session import _start_session

    return _start_session(
        api_key,
        root_url=root_url,
        host=host,
        debug=debug,
        strict=strict,
        use_keyring=use_keyring,
        log_level=log_level,
    )




import argparse
import code
import readline
import rlcompleter  # noqa  provides python shell autocomplete
import dli
import signal
import sys

from distutils.util import strtobool
from dli import connect

try:
    import IPython  # noqa
    from traitlets.config.loader import Config  # noqa
except ImportError:
    IPython = None
    Config = None


def aws(dli_client):
    try:
        dli_client.aws_authenticator.start()
        dli_client.aws_authenticator.join()
    except (KeyboardInterrupt, SystemExit):
        print('\n')
        dli_client.logger.info('Exiting')
    except Exception:
        sys.exit(1)


def shell(dli_client, basic_shell=False):
    banner = (
        f'DLI {dli.__version__} session started. '
        'Can be used by accessing `client`\n'
    )

    if IPython and not basic_shell:
        config = Config()
        config.TerminalInteractiveShell.banner2 = banner
        IPython.start_ipython(
            argv=[], user_ns={'client': dli_client},
            config=config
        )
    else:
        readline.parse_and_bind('tab:complete')
        console = code.InteractiveConsole(locals={
            'client': dli_client
        })
        console.interact(
            banner=banner
        )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--root_url', default=None,
        type=str,
    )

    parser.add_argument(
        '--host', default=None,
        type=str,
    )

    parser.add_argument(
        '--debug', default=None,
        type=strtobool,
    )

    parser.add_argument(
        '--strict', default=None,
        type=strtobool,
    )

    parser.add_argument(
        '--use_keyring', default=None,
        type=strtobool,
    )

    parser.add_argument(
        '--log_level', default=None,
        type=str,
    )

    subparser = parser.add_subparsers(dest='subcommand')
    subparser.add_parser(
        'aws',
        help=(
            'Runs a process which writes to ~/.aws/credentials. Keeping it '
            'synced with the credentials provided by the DLI.'
        )
    )

    shell_parser = subparser.add_parser(
        'shell',
        help=(
            'Runs the DLI in a Python shell. (IPython if available).'
        )
    )

    shell_parser.add_argument(
        '--basic', action='store_true',
        help='Use basic Python shell without IPython',
        default=False
    )

    subparser.add_parser(
        'version',
        help=(
            'Displays the version'
        )
    )

    subparser.required = True
    args = parser.parse_args()

    connection_args = {
        k:v for k,v in vars(args).items() if v is not None
    }

    basic_shell = connection_args.pop('basic', False)
    subcommand = connection_args.pop('subcommand')

    dli_client = connect(**connection_args)

    if subcommand == 'aws':
        aws(dli_client)

    if subcommand == 'shell':
        shell(dli_client, basic_shell)

    if subcommand == 'version':
        print(f'{dli.__version__}')

    sys.exit(0)
